from abc import ABC
from typing import Type, Union, Callable

from scipy.stats import rv_continuous, rv_discrete

import numpy as np

from snake_eyes import Distribution
from snake_eyes.buffered_distribution import BuffererMakerDistribution
from snake_eyes.bufferer import Bufferer
from snake_eyes.support_space import ContinuousSupportSpace

RV = Union[rv_discrete, rv_continuous]


def rv_bufferer_class(rv: RV):
    """
    Create a Bufferer subclass that wraps a scipy rv
    """
    class Ret(Bufferer):
        def __init__(self, *args, **kwargs):
            super().__init__()
            self.args = args
            self.kwargs = kwargs

        def get_buffer(self, size) -> np.ndarray:
            return rv.rvs(*self.args, size=size, **self.kwargs)

    return Ret


class RVAdapterBase(Distribution, ABC):
    """
    an mixin subclass that implements many methods with a frozen scipy.rv instance
    """
    def __init__(self, frozen):
        self.frozen = frozen

    def mean(self):
        return self.frozen.mean()

    def variance(self):
        return self.frozen.var()

    def deviation(self):
        return self.frozen.std()

    def cumulative_density(self, k):
        return self.frozen.cdf(k)

    def sample_mean(self, sample_size=256, abs_tol=0.0, rel_tol=1e-7, func: Callable[[float], float] = None):
        if func is None:
            self.mean()
        return self.frozen.expect(func)


def _rv_class(rv: RV):
    """
    create a common superclass for a generic rv
    """
    bufferer_cls = rv_bufferer_class(rv)

    class Ret(BuffererMakerDistribution[float], RVAdapterBase):
        def __init__(self, *args, **kwargs):
            BuffererMakerDistribution.__init__(self, bufferer_cls, args, kwargs)
            RVAdapterBase.__init__(self, rv(*args, **kwargs))

    return Ret


def rv_continuous_class(rv: rv_continuous, name: str = None, module: str = None, **kwargs) \
        -> Type[BuffererMakerDistribution[float]]:
    """
    create a distribution for a scipy.rv_continuous
    :param rv: the rv_continuous instance to wrap
    :param name: if set, sets the name of the class
    :param module: if set, sets the module of the class
    :param kwargs: additional attributes of the class to set
    """
    base = _rv_class(rv)

    class Ret(base):
        def probability(self, k):
            return self.frozen.pdf(k)

        def support_space(self):
            a, b = self.frozen.interval(1)
            return ContinuousSupportSpace(a, b)

    for k, v in kwargs.items():
        setattr(Ret, k, v)

    if module:
        Ret.__module__ = module
    if name:
        Ret.__qualname__ = Ret.__name__ = name

    return Ret


def rv_discrete_class(rv: rv_discrete, name: str = None, module: str = None, support_space_cls=None, **kwargs) \
        -> Type[BuffererMakerDistribution[float]]:
    """
    create a distribution for a scipy.rv_discrete
    :param rv: the rv_discrete instance to wrap
    :param name: if set, sets the name of the class
    :param module: if set, sets the module of the class
    :param support_space_cls: the class for the support space, called with the minimum and maximum of the distribution
    :param kwargs: additional attributes of the class to set
    """
    base = _rv_class(rv)

    class Ret(base):
        def probability(self, k):
            return self.frozen.pmf(k)

        def support_space(self):
            if support_space_cls:
                a, b = self.frozen.interval(1)
                # todo there is currently a bug in scipy that intervals for discrete values are offset by 1 on
                #  lower bound, fix
                return support_space_cls(a + 1, b)
            return None

    for k, v in kwargs.items():
        setattr(Ret, k, v)

    if module:
        Ret.__module__ = module
    if name:
        Ret.__qualname__ = Ret.__name__ = name

    return Ret



from abc import abstractmethod, ABC
from heapq import heapify, heappop
from typing import Generic, TypeVar, Optional

import numpy as np
from scipy.special import binom

from snake_eyes import DistributionKind
from snake_eyes.buffered_distribution import BufferedDistribution
from snake_eyes.bufferer import Bufferer
from snake_eyes.distribution import Distribution
from snake_eyes.support_space import SupportSpace, SumSupportSpace, ProdSupportSpace
from snake_eyes.util import prod

T = TypeVar('T')


class Drop(Distribution[T], Generic[T]):
    def __init__(self, inner: Distribution[T], total_rolls: int, drop_low: int = 0, drop_high: int = 0):
        self.inner = inner
        self.total_rolls = total_rolls
        self.drop_low = drop_low
        self.drop_high = drop_high

    @classmethod
    @abstractmethod
    def func(cls, parts):
        pass

    def get(self) -> T:
        heap = self.inner.get_n(self.total_rolls)
        if not isinstance(heap, list):
            heap = list(heap)
        heapify(heap)
        for _ in range(self.drop_low):
            heappop(heap)
        ret = []
        for _ in range(self.actual_parts()):
            ret.append(heappop(heap))
        return self.func(ret)

    def actual_parts(self):
        return self.total_rolls - self.drop_low - self.drop_high

    def cumulative_density(self, k) -> Optional[float]:
        if self.actual_parts() == 1:
            i_cdf = self.inner.cumulative_density(k)
            if i_cdf is None:
                return None
            return sum(
                binom(self.total_rolls, i) * i_cdf ** i * (1 - i_cdf) ** (T - i)
                for i in range(self.drop_low + 1, self.total_rolls + 1)
            )
        return super().cumulative_density(k)

    def probability(self, k) -> Optional[float]:
        if self.actual_parts() == 1:
            ss = self.inner.support_space()
            if ss is None:
                return None
            kind = ss.kind()
            if kind is None:
                return None
            i_cdf = self.inner.cumulative_density(k)
            i_pdf = self.inner.probability(k)
            if i_cdf is None or i_pdf is None:
                return None
            if kind == DistributionKind.Continuous:
                return (self.drop_low + 1) * i_pdf \
                       * binom(self.total_rolls, (self.drop_low + 1)) \
                       * (i_cdf ** self.drop_low) \
                       * (1 - i_cdf) ** (T - self.drop_low - 1)
            elif kind == DistributionKind.Discrete:
                sum(
                    binom(self.total_rolls, i) *
                    ((i_cdf ** i * (1 - i_cdf) ** (T - i))
                     - ((i_cdf - i_pdf) ** i * (1 - (i_cdf - i_pdf)) ** (T - i)))
                    for i in range(self.drop_low + 1, self.total_rolls + 1)
                )
            else:
                return None
        return super().probability(k)

    def __hash__(self):
        return hash((type(self), self.inner, self.drop_high, self.drop_low, self.total_rolls))


class SumDrop(Drop[T], Generic[T]):
    @classmethod
    def func(cls, parts):
        return sum(parts)

    def support_space(self) -> Optional[SupportSpace]:
        i_ss = self.inner.support_space()
        if i_ss is None:
            return None
        return SumSupportSpace((i_ss,) * self.actual_parts())


class ProdDrop(Drop[T], Generic[T]):
    @classmethod
    def func(cls, parts):
        return prod(parts)

    def support_space(self) -> Optional[SupportSpace]:
        i_ss = self.inner.support_space()
        if i_ss is None:
            return None
        return ProdSupportSpace((i_ss,) * self.actual_parts())


class DropBufferer(Bufferer, ABC):
    def __init__(self, inner: Bufferer, total_rolls: int, drop_low: int = 0, drop_high: int = 0):
        super().__init__()
        self.inner = inner
        self.total_rolls = total_rolls
        self.drop_low = drop_low
        self.drop_high = drop_high

    @classmethod
    @abstractmethod
    def func(cls, heap):
        pass

    def get_buffer(self, size) -> np.ndarray:
        heap = self.inner.get_buffer((size, self.total_rolls))
        if self.drop_low > 0:
            heap.partition(self.drop_low - 1, axis=1)
            heap = heap[:, self.drop_low:]
        if self.drop_high > 0:
            ind = len(heap) - self.drop_high
            heap.partition(ind, axis=1)
            heap = heap[:, :ind]
        return self.func(heap)


class DropSumBufferer(DropBufferer):
    @classmethod
    def func(cls, heap):
        return np.sum(heap, axis=1)


class DropProdBufferer(DropBufferer):
    @classmethod
    def func(cls, heap):
        return np.prod(heap, axis=1)


class BufferedSumDrop(BufferedDistribution[T], SumDrop[T], Generic[T]):
    def __init__(self, inner: BufferedDistribution[T], total_rolls: int, drop_low: int = 0, drop_high: int = 0):
        SumDrop.__init__(self, inner, total_rolls, drop_low, drop_high)
        BufferedDistribution.__init__(self, DropSumBufferer(inner.bufferer, total_rolls, drop_low, drop_high))


class BufferedProdDrop(BufferedDistribution[T], ProdDrop[T], Generic[T]):
    def __init__(self, inner: BufferedDistribution[T], total_rolls: int, drop_low: int = 0, drop_high: int = 0):
        ProdDrop.__init__(self, inner, total_rolls, drop_low, drop_high)
        BufferedDistribution.__init__(self, DropSumBufferer(inner.bufferer, total_rolls, drop_low, drop_high))


def drop_sum(inner: Distribution[T], total_rolls: int, drop_low: int = 0, drop_high: int = 0) -> Distribution[T]:
    if drop_low == 0 and drop_high == 0:
        return inner
    if (drop_low + drop_high) >= total_rolls:
        raise ValueError('cannot drop all rolls')
    if isinstance(inner, BufferedDistribution):
        return BufferedSumDrop(inner, total_rolls, drop_low, drop_high)
    return SumDrop(inner, total_rolls, drop_low, drop_high)


def drop_prod(inner: Distribution[T], total_rolls: int, drop_low: int = 0, drop_high: int = 0) -> Distribution[T]:
    if drop_low == 0 and drop_high == 0:
        return inner
    if (drop_low + drop_high) >= total_rolls:
        raise ValueError('cannot drop all rolls')
    if isinstance(inner, BufferedDistribution):
        return BufferedProdDrop(inner, total_rolls, drop_low, drop_high)
    return ProdDrop(inner, total_rolls, drop_low, drop_high)



from __future__ import annotations

from abc import abstractmethod
from itertools import product
from math import sqrt
from numbers import Number, Real
from typing import Generic, TypeVar, Optional, Iterable, Union, Iterator, Any, Callable, Sequence

import numpy as np
from dyndis import MultiDispatch, Self

from snake_eyes.support_space import SupportSpace, DistributionKind, DiscreteFiniteSupportSpace, SumSupportSpace, \
    ProdSupportSpace, ReciprocalSupportSpace, ZipSupportSpace
from snake_eyes.util import prod, _maybe_parenthesise

add = MultiDispatch('__add__')
mul = MultiDispatch('__mul__')
div = MultiDispatch('__truediv__')
zip_ = MultiDispatch('zip')

T = TypeVar('T')


class Distribution(Iterator, Generic[T]):
    """
    A fully independent random variable
    """

    @staticmethod
    def coerce(x):
        """
        :return: x if it is a Distribution, or wrap it in a ConstDistribution otherwise
        """
        if isinstance(x, Distribution):
            return x
        return ConstDistribution(x)

    __radd__ = __add__ = add.op()
    __rmul__ = __mul__ = mul.op()
    __truediv__ = div.op()

    zip = zip_.method()

    @abstractmethod
    def get(self) -> T:
        """
        :return: roll a value from the distribution and return it
        """
        pass

    def get_n(self, n) -> Sequence[T]:
        """
        :return: roll n values from the distribution and return them as a sequence
        """
        yield from tuple(self.get() for _ in range(n))

    def mean(self) -> Optional[T]:
        """
        :return: the expected value of the distribution, or None if it cannot be exactly calculated
        :note: if the distribution has a finite support space and supports probability calls, this function
         has a default implementation
        """
        ss = self.support_space()
        if not ss or not ss.is_finite():
            return None
        ret = 0
        for x_i in ss:
            prob = self.probability(x_i)
            if prob is None:
                return None
            ret += prob * x_i
        return ret

    def variance(self) -> Optional[T]:
        """
        :return: the variance of the distribution, or None if it cannot be exactly calculated
        :note: if the distribution has a finite support space and supports probability calls, this function
         has a default implementation
        """
        ss = self.support_space()
        if not ss or not ss.is_finite():
            return None
        mean = 0
        sq_mean = 0
        for x_i in ss:
            prob = self.probability(x_i)
            if prob is None:
                return None
            pxi = prob * x_i
            mean += pxi
            sq_mean += pxi * x_i
        return sq_mean - mean ** 2

    def support_space(self) -> Optional[SupportSpace]:
        """
        :return: the support space of the distribution, or None if it cannot be exactly calculated
        """
        return None

    def cumulative_density(self, k) -> Optional[float]:
        """
        :return: the probability of the distribution returning a value less than or equal to k,
         or None if it cannot be exactly calculated
        :note: if the distribution has a finite support space and supports probability calls, this function
         has a default implementation
        """
        ss = self.support_space()
        if not ss:
            return None
        if k >= ss.maximum():
            return 1
        if k < ss.minimum():
            return 0
        if not ss.is_finite():
            return None
        ret = 0
        for x_i in ss:
            if x_i > k:
                continue
            prob = self.probability(x_i)
            if prob is None:
                return None
            ret += prob * x_i
        return ret

    def probability(self, k) -> Optional[float]:
        """
        :return: the probability of the distribution returning a value equal to k,
         or None if it cannot be exactly calculated. For continuous distributions, return the pmf of the distribution
        :note: if the distribution has a finite support space and supports probability calls, this function
         has a default implementation
        """
        ss = self.support_space()
        if not ss:
            return None
        if ss.contains(k) is False:
            return 0
        return None

    def deviation(self) -> Optional[T]:
        """
        :return: the support space of the distribution, or None if it cannot be exactly calculated
        """
        v = self.variance()
        if v is None:
            return v
        return sqrt(v)

    def __neg__(self) -> Distribution:
        """
        :return: a distribution that returns the negatives of the original distribution
        """
        return -1 * self

    def __sub__(self, other):
        return self + (-other)

    def __rsub__(self, other):
        return other + (-self)

    def truncate(self, min=None, max=None):
        """
        :param min: The lower bound, or None
        :param max: The upper bound, or None
        :return: a truncated distribution, based on self and bound between min and max
        """
        if min is None and max is None:
            return self
        ss = self.support_space()
        if ss:
            a = ss.minimum()
            b = ss.maximum()
            if (min is None or a >= min) and (max is None or b <= max):
                return self
        return TruncatedDistribution(self, min, max)

    def reciprocal(self):
        """
        :return: a distribution reciprocal to this one
        """
        return ReciprocalDistribution(self)

    @div.implementor(-1)
    def truediv(self: Union[Self, Number], other):
        return self * other.reciprocal()

    @div.implementor(-1)
    def truediv(self, other: Real):
        return self * (1 / other)

    def __rtruediv__(self, other):
        return other * self.reciprocal()

    @add.implementor(priority=1, symmetric=True)
    def add(self, other: Number):
        if other == 0:
            return self
        return NotImplemented

    @add.implementor(priority=-1, symmetric=True)
    def add(self, other: object):
        return self + Distribution.coerce(other)

    @mul.implementor(priority=1, symmetric=True)
    def mul(self, other: Number):
        if other == 1:
            return self
        if other == 0:
            return self.coerce(other)
        return NotImplemented

    @mul.implementor(priority=-1, symmetric=True)
    def mul(self, other: object):
        return self * Distribution.coerce(other)

    @zip_.implementor()
    def zip_(self, other):
        return ZipDistribution((self, other))

    def approx_mean(self, **kwargs):
        """
        :param kwargs: forwarded to sample_mean, if necessary
        :return: the mean of the distribution, either exact or approximated using sample_mean
        """
        if 'func' not in kwargs:
            exact = self.mean()
            if exact is not None:
                return exact
        return self.sample_mean(**kwargs)

    def approx_variance(self, **kwargs):
        """
        :param kwargs: forwarded to sample_variance, if necessary
        :return: the variance of the distribution, either exact or approximated using sample_variance
        """
        exact = self.variance()
        if exact is not None:
            return exact
        return self.sample_variance(**kwargs)

    def approx_deviation(self, **kwargs):
        """
        :param kwargs: forwarded to sample_variance, if necessary
        :return: the standard deviation of the distribution, either exact or approximated using sample_variance
        """
        exact = self.deviation()
        if exact is not None:
            return exact
        return sqrt(self.approx_variance(**kwargs))

    def approx_cumulative_density(self, k, **kwargs):
        """
        :param kwargs: forwarded to sample_cumulative_density, if necessary
        :return: the cdf of the distribution, either exact or approximated using sample_cumulative_density
        """
        exact = self.cumulative_density(k)
        if exact is not None:
            return exact
        return self.sample_cumulative_density(k, **kwargs)

    def approx_probability(self, k, **kwargs):
        """
        :param kwargs: forwarded to sample_probability, if necessary
        :return: the probability of the distribution, either exact or approximated using sample_probability
        """
        exact = self.probability(k)
        if exact is not None:
            return exact
        return self.sample_probability(k, **kwargs)

    def sample_mean(self, sample_size=4096, func: Callable[[T], T] = None):
        """
        :param sample_size: the size of the sample
        :param func: a function to apply to the members (if set, calculates E[func(X)]
        :return: an approximate mean of the distribution calculated from a sample
        """
        batch = self.get_n(sample_size)
        if isinstance(batch, np.ndarray):
            if func:
                batch = func(batch)
            ret_s = np.sum(batch)
        else:
            if func:
                batch = (func(b) for b in batch)
            ret_s = sum(batch)
        return ret_s / sample_size

    def sample_variance(self, sample_size=4096):
        """
        :param sample_size: the size of the sample
        :return: an approximate variance of the distribution calculated from a sample
        """
        mean = self.approx_mean(sample_size=sample_size)
        return self.sample_mean(func=lambda x: x ** 2, sample_size=sample_size) - mean ** 2

    def sample_cumulative_density(self, k, sample_size=4096):
        """
        :param sample_size: the size of the sample
        :return: an approximate cdf of the distribution calculated from a sample
        """
        return self.sample_mean(func=lambda x: x <= k, sample_size=sample_size)

    def sample_probability(self, k, sample_size=4096, epsilon=None):
        """
        :param sample_size: the size of the sample
        :param epsilon: for continuous distributions, the epsilon of the approximation range.
         By default, uses the support space to generate an epsilon
        :return: an approximate pdf or pmf of the distribution calculated from a sample
        """
        ss = self.support_space()
        data = self.get_n(sample_size)
        if ss and ss.kind() == DistributionKind.Continuous:
            if epsilon is None:
                # by default assuming uniform, sqrt(sample_size) will be within epsilon
                epsilon = (ss.maximum() - ss.minimum()) / sqrt(sample_size)
            if isinstance(data, np.ndarray):
                total = np.sum((data <= (k + epsilon)) & ((k - epsilon) <= data))
            else:
                total = sum(1 for i in data if k - epsilon <= i <= k + epsilon)
            return total / (2 * epsilon * sample_size)
        else:
            if isinstance(data, np.ndarray):
                total = np.sum(data == k)
            else:
                total = sum(1 for i in data if i == k)
            return total / sample_size

    def __next__(self):
        """
        :return: equal to get()
        """
        return self.get()

    def __rmatmul__(self, other: int) -> Distribution:
        if other == 1:
            return self
        if other == 0:
            return NotImplemented
        if other < 0:
            return -(-other @ self)
        return sum(self for _ in range(other))

    def __lt__(self, other: T) -> Optional[float]:
        ret = self.cumulative_density(other)
        if ret is None:
            return None
        s = self.support_space()
        if s is not None and s.kind() != DistributionKind.Continuous:
            p = self.probability(other)
            if p is None:
                return None
            ret -= p
        return ret

    def __le__(self, other: T) -> Optional[float]:
        return self.cumulative_density(other)

    def __gt__(self, other: T) -> Optional[float]:
        return 1 - self.cumulative_density(other)

    def __ge__(self, other: T) -> Optional[float]:
        ret = 1 - self.cumulative_density(other)
        if ret is None:
            return None
        s = self.support_space()
        if s is not None and s.kind() != DistributionKind.Continuous:
            p = self.probability(other)
            if p is None:
                return None
            ret += p
        return ret

    def map(self, func: Callable[[T], F]) -> Distribution[F]:
        return MappedDistribution(self, func)


class ConstDistribution(Distribution[T], Generic[T]):
    """
    A distribution that always returns a constant value
    """

    def __init__(self, value):
        self.value = value

    def get(self):
        return self.value

    def get_n(self, n):
        return np.full(shape=n, fill_value=n)

    def mean(self):
        return self.value

    def variance(self):
        return 0

    def support_space(self):
        return DiscreteFiniteSupportSpace((self.value,), True)

    def cumulative_density(self, k):
        return int(k >= self.value)

    def probability(self, k):
        return int(k == self.value)

    def reciprocal(self):
        return type(self)(1 / self.value)

    def truncate(self, min=None, max=None):
        if (min is not None and min > self.value) \
                or (max is not None and max < self.value):
            raise ValueError("can't truncate a constant distribution outside its value")
        return self

    @add.implementor(symmetric=True)
    def add(self, other: Distribution):
        return self.value + other

    @mul.implementor(symmetric=True)
    def mul(self, other: Distribution):
        return self.value * other

    @add.implementor()
    def add(self, other: Self):
        return type(self)(self.value + other.value)

    @add.implementor(symmetric=True)
    def add(self, other: Number):
        return type(self)(self.value + other)

    @mul.implementor()
    def mul(self, other: Self):
        return type(self)(self.value * other.value)

    @mul.implementor(symmetric=True)
    def mul(self, other: Number):
        return type(self)(self.value * other)

    @div.implementor()
    def truediv(self, other):
        return type(self)(self.value / other.value)

    @div.implementor()
    def truediv(self, other: Number):
        return type(self)(self.value / other)

    @div.implementor()
    def truediv(self: Real, other):
        return type(self)(self / other.value)

    def __eq__(self, other):
        return type(self) is type(other) and self.value == other.value

    def __repr__(self):
        return type(self).__name__ + f"({self.value})"

    def __hash__(self):
        return hash(self.value)


class SymmetricOpDistribution(Distribution[T], Generic[T]):
    """
    Abstract class of composite distributions with a summetric and invertible function to compose them
    """

    def __init__(self, parts: Iterable[Distribution[T]]):
        parts = tuple(parts)
        assert len(parts) >= 2
        self.parts = parts

    @classmethod
    @abstractmethod
    def func(cls, parts):
        """
        combine multiple outputs of sub-distributions
        """
        pass

    def get(self):
        return self.func(p.get() for p in self.parts)

    def get_n(self, n):
        n_parts = tuple(p.get_n(n) for p in self.parts)
        for n_p in zip(*n_parts):
            yield self.func(n_p)

    @classmethod
    @abstractmethod
    def rev_func(cls, target, parts):
        """
        reverse of func, return None if no inverse exists
        :note: it should always be true that cls.func([cls.rev_func(target,x), *x]) == target
        """
        pass

    def probability(self, k):
        finite_parts = []
        finite_spaces = []
        other_parts = []
        for p in self.parts:
            ss = p.support_space()
            if ss is None or not ss.is_finite():
                other_parts.append(p)
                if len(other_parts) >= 2:
                    return None
            else:
                finite_spaces.append(ss)
                finite_parts.append(p)
        assert finite_parts
        if not other_parts:
            other_parts = finite_parts[-1:]
            del finite_parts[-1]
            del finite_spaces[-1]
        other, = other_parts
        total = 0
        for possibility in product(*finite_spaces):
            probs = tuple(part.probability(poss) for (part, poss) in zip(finite_parts, possibility))
            if None in probs:
                return None
            poss_prob = prod(probs)
            if poss_prob == 0:
                continue
            req = self.rev_func(k, possibility)
            if req is None:
                continue
            other_prob = other.probability(req)
            if other_prob is None:
                return None
            total += poss_prob * other_prob
        return total

    def sample_probability(self, k, **kwargs):
        discrete_finite_parts = []
        discrete_support_spaces = []
        other_parts = []
        for p in self.parts:
            ss = p.support_space()
            if ss is None or not ss.is_finite():
                other_parts.append(p)
            else:
                discrete_support_spaces.append(ss)
                discrete_finite_parts.append(p)

        if not discrete_finite_parts:
            return super().sample_probability(k, **kwargs)

        if not other_parts:
            other_parts = discrete_finite_parts[-1:]
            del discrete_finite_parts[-1]
            del discrete_support_spaces[-1]

        other = self.func(other_parts)
        total = 0

        for possibility in product(*discrete_support_spaces):
            probs = tuple(
                part.approx_probability(poss, **kwargs) for (part, poss) in zip(discrete_finite_parts, possibility)
            )
            if None in probs:
                return None
            poss_prob = prod(probs)
            if poss_prob == 0:
                continue
            req = self.rev_func(k, possibility)
            if req is None:
                continue
            other_prob = other.approx_probability(req, **kwargs)
            if other_prob is None:
                return None
            total += poss_prob * other_prob
        return total

    def __hash__(self):
        return hash((type(self), frozenset(self.parts)))

    def __eq__(self, other):
        return type(self) is type(other) and set(self.parts) == set(other.parts)


class SumDistribution(SymmetricOpDistribution[T], Generic[T]):
    """
    A distribution for a sum of other distributions
    """

    def mean(self):
        ret = 0
        for p in self.parts:
            m = p.mean()
            if m is None:
                return None
            ret += m
        return ret

    def variance(self):
        ret = 0
        for p in self.parts:
            m = p.variance()
            if m is None:
                return None
            ret += m
        return ret

    def support_space(self):
        spaces = []
        for p in self.parts:
            ss = p.support_space()
            if ss is None:
                return None
            spaces.append(ss)
        return SumSupportSpace(spaces)

    @classmethod
    def func(cls, parts):
        return sum(parts)

    @classmethod
    def rev_func(cls, target, parts):
        return target - sum(parts)

    @add.implementor(symmetric=True)
    def add(self, other: Distribution):
        p = self.parts + (other,)
        return type(self)(p)

    @mul.implementor(symmetric=True)
    def mul(self, other: Any):
        return sum(p * other for p in self.parts)

    @div.implementor()
    def truediv(self, other: Any):
        return sum(p / other for p in self.parts)

    def __str__(self):
        return " + ".join(_maybe_parenthesise(p) for p in self.parts)


class ProductDistribution(SymmetricOpDistribution[T], Generic[T]):
    """
    A distribution for a product of other distributions
    """

    @classmethod
    def func(cls, parts):
        return prod(parts)

    @classmethod
    def rev_func(cls, target, parts):
        if any(p == 0 for p in parts):
            return None
        return target / prod(parts)

    def mean(self):
        ret = 0
        for p in self.parts:
            m = p.mean()
            if m is None:
                return None
            ret *= m
        return ret

    def variance(self):
        variances = tuple(p.variance() for p in self.parts)
        means = tuple(p.mean() for p in self.parts)
        if any(d is None for d in variances) or any(m is None for m in means):
            return None
        sq_means = tuple(m * m for m in means)
        pos = zip(variances, sq_means)
        return sum(
            prod(x) for x in product(*pos)
        ) - prod(sq_means)

    def support_space(self):
        spaces = []
        for p in self.parts:
            ss = p.support_space()
            if ss is None:
                return None
            spaces.append(ss)
        return ProdSupportSpace(spaces)

    @mul.implementor(symmetric=True)
    def mul(self, other: Distribution):
        p = (*self.parts, other)
        return type(self)(p)

    def __str__(self):
        return " * ".join(_maybe_parenthesise(p) for p in self.parts)


class ReciprocalDistribution(Distribution[T], Generic[T]):
    """
    A distribution for a reciprocal of another distribution
    """

    def __init__(self, inner: Distribution):
        self.inner = inner
        ss = self.inner.support_space()
        if ss and ss.contains(0):
            raise ZeroDivisionError("can't get inverse of distribution that might return 0")

    def get(self):
        return 1 / self.inner.get()

    def get_n(self, n):
        yield from (1 / i for i in self.inner.get_n(n))

    def reciprocal(self):
        return self.inner

    def support_space(self):
        iss = self.inner.support_space()
        return iss and ReciprocalSupportSpace(iss)

    def probability(self, k):
        return self.inner.probability(1 / k)

    def __eq__(self, other):
        return type(self) is type(other) and set(self.inner) == set(other.inner)

    def __str__(self):
        return f"1/" + _maybe_parenthesise(self.inner)

    def __hash__(self):
        return hash((type(self), self.inner))


class ZipDistribution(Distribution[tuple]):
    """
    A distribution for a zipping of other distributions
    """

    def __init__(self, parts: Iterable[Distribution]):
        self.parts = tuple(parts)

    def get(self) -> T:
        return tuple(p.get() for p in self.parts)

    def get_n(self, n) -> T:
        n_parts = tuple(p.get_n(n) for p in self.parts)
        yield from zip(*n_parts)

    def mean(self) -> Optional[T]:
        ret = []
        for p in self.parts:
            m = p.mean()
            if m is None:
                return None
            ret.append(m)
        return ret

    def variance(self) -> Optional[T]:
        ret = []
        for p in self.parts:
            m = p.variance()
            if m is None:
                return None
            ret.append(m)
        return ret

    def deviation(self) -> Optional[T]:
        ret = []
        for p in self.parts:
            m = p.deviation()
            if m is None:
                return None
            ret.append(m)
        return ret

    def support_space(self) -> Optional[SupportSpace]:
        parts = []
        for p in self.parts:
            ss = p.support_space()
            if ss is None:
                return None
            parts.append(ss)
        return ZipSupportSpace(parts)

    def probability(self, k):
        if len(k) != len(self.parts):
            return 0
        ret = 1
        for (p, i) in zip(self.parts, k):
            m = p.probability(i)
            if m is None:
                return None
            if m == 0:
                return 0
            ret *= m
        return ret

    def sample_probability(self, k, **kwargs):
        if len(k) != len(self.parts):
            return 0
        ret = 1
        p: Distribution
        for (p, i) in zip(self.parts, k):
            m = p.approx_probability(i)
            if m is None:
                return None
            if m == 0:
                return 0
            ret *= m
        return ret

    @zip_.implementor()
    def zip_(self, other: Distribution):
        return ZipDistribution((*self.parts, other))

    @zip_.implementor()
    def zip_(self, other: Self):
        return ZipDistribution((*self.parts, *other.parts))

    @zip_.implementor()
    def zip_(self: Distribution, other: Self):
        return ZipDistribution((self, *other.parts))

    def __eq__(self, other):
        return type(self) is type(other) and self.parts == other.parts

    def __repr__(self):
        return type(self).__name__ + "(" + ", ".join(_maybe_parenthesise(p) for p in self.parts) + ")"

    def __hash__(self):
        return hash((type(self), self.parts))


class TruncatedDistribution(Distribution[T], Generic[T]):
    """
    A truncated wrapper for a distribution, re-rolling all results outsize the bounds
    """

    def __init__(self, inner: Distribution[T], min=None, max=None):
        self.inner = inner
        self.min = min
        self.max = max

    def get(self) -> T:
        for ret in self.inner:
            if self.min is not None and self.min > ret:
                continue
            if self.max is not None and self.max < ret:
                continue
            return ret

    def get_n(self, n) -> T:
        density = self._valid_density()
        if density is None:
            density = 1
        ret = []
        while len(ret) < n:
            buffer = (
                i for i in self.inner.get_n(int((n - len(ret)) / density))
                if (self.min is None or i >= self.min) and (self.max is None or i <= self.max)
            )
            ret.extend(buffer)
        if len(ret) >= n:
            del ret[n:]
        return ret

    def __neg__(self):
        n = -self.min if self.min else self.min
        x = -self.max if self.max else self.max
        return (-self.inner).truncate(x, n)

    @add.implementor(symmetric=True)
    def add(self, other: Number):
        n = self.min + other if self.min else self.min
        x = self.max + other if self.max else self.max
        return (self.inner + other).truncate(n, x)

    @mul.implementor(symmetric=True)
    def mul(self, other: Real):
        n = self.min * other if self.min else self.min
        x = self.max * other if self.max else self.max
        if other < 0:
            n, x = x, n
        return (self.inner * other).truncate(n, x)

    def truncate(self, min=None, max=None):
        if min is None or (self.min is not None and self.min > min):
            min = self.min

        if max is None or (self.max is not None and self.max < max):
            max = self.max

        if (min, max) == (self.min, self.max):
            return self

        return self.inner.truncate(min, max)

    def _valid_density(self):
        if self.max is not None:
            m = self.inner.cumulative_density(self.max)
            if m is None:
                return None
            ret = m
        else:
            ret = 1

        if self.min is not None:
            m = self.inner.cumulative_density(self.min)
            if m is None:
                return None
            ret -= m
        return ret

    def support_space(self):
        iss = self.inner.support_space()
        return iss and iss.truncate(self.min, self.max)

    def cumulative_density(self, k):
        vd = self._valid_density()
        if vd is None:
            return None
        f_a = 0
        if self.min is not None:
            f_a = self.inner.cumulative_density(self.min)
            if f_a is None:
                return None
        return (self.inner.cumulative_density(k) - f_a) / vd

    def probability(self, k):
        vd = self._valid_density()
        if vd is None:
            return None
        return self.inner.probability(k) / vd

    def __eq__(self, other):
        return type(self) is type(other) and (self.inner, self.min, self.max) == (other.inner, other.min, other.max)

    def __hash__(self):
        return hash((type(self), self.inner, self.min, self.max))


F = TypeVar('F')


class MappedDistribution(Distribution[T], Generic[F, T]):
    def __init__(self, inner: Distribution[F], func: Callable[[F], T]):
        self.inner = inner
        self.func = func

    def get(self) -> T:
        return self.func(self.inner.get())

    def get_n(self, n) -> Sequence[T]:
        return [self.func(i) for i in self.inner.get_n(n)]

    def __repr__(self):
        return f'{self.inner!r}.map({self.func!r})'

    def __eq__(self, other):
        return type(self) is type(other) and self.inner == other.inner and self.func == other.func

    def __hash__(self):
        return hash((type(self), self.inner, self.func))



from math import floor, ceil

from scipy.stats import uniform, randint, norm, truncnorm, bernoulli, geom, hypergeom, poisson, zipf, triang

from snake_eyes.distribution import add, mul
from snake_eyes.from_rv import rv_continuous_class, rv_discrete_class
from snake_eyes.support_space import DiscreteWholeSupportSpace


class UniformDistribution(rv_continuous_class(uniform)):
    def __init__(self, a, b):
        self.a = a
        self.b = b
        super().__init__(a, b - a)

    @add.implementor(symmetric=True)
    def add(self, other: float):
        return type(self)(self.a + other, self.b + other)

    @mul.implementor(symmetric=True)
    def mul(self, other: float):
        a = self.a
        b = self.b
        if other < 0:
            a, b = b, a
        return type(self)(a * other, b * other)

    def truncate(self, a=None, b=None):
        a = max(a, self.a) if a is not None else self.a
        b = min(b, self.b) if b is not None else self.b
        if (a, b) == (self.a, self.b):
            return self
        return type(self)(a, b)

    def args_str(self):
        return self.a, self.b


class UniformDiscreteDistribution(rv_discrete_class(randint, support_space_cls=DiscreteWholeSupportSpace)):
    def __init__(self, a, b):
        self.a = a
        self.b = b
        super().__init__(a, b + 1)

    def __neg__(self):
        return type(self)(-self.b, -self.a)

    @add.implementor(symmetric=True)
    def add(self, other: int):
        return type(self)(self.a + other, self.b + other)

    @mul.implementor(symmetric=True)
    def mul(self, other: float):
        if other != -1:
            return NotImplemented
        return -self

    def truncate(self, a=None, b=None):
        a = max(ceil(a), self.a) if a is not None else self.a
        b = min(floor(b), self.b) if b is not None else self.b
        if (a, b) == (self.a, self.b):
            return self
        return type(self)(a, b)

    def args_str(self):
        return self.a, self.b


class NormalDistribution(rv_continuous_class(norm)):
    def __init__(self, mu, sd):
        super().__init__(mu, sd)
        self.mu = mu
        self.sd = sd

    @add.implementor(symmetric=True)
    def add(self, other: float):
        return type(self)(self.mu + other, self.sd)

    @mul.implementor(symmetric=True)
    def mul(self, other: float):
        return type(self)(self.mu * other, self.sd * other)

    def truncate(self, min=None, max=None):
        if min is not None and max is not None:
            return TruncatedNormalDistribution(self.mu, self.sd, min, max)
        return super().truncate(min, max)


class TruncatedNormalDistribution(rv_continuous_class(truncnorm)):
    def __init__(self, mu, sd, a, b):
        self.a_adjusted = (a - mu) / sd
        self.b_adjusted = (b - mu) / sd
        super().__init__(mu, sd, self.a_adjusted, self.b_adjusted)
        self.mu = mu
        self.sd = sd
        self.a = a
        self.b = b

    def _inner(self):
        return NormalDistribution(self.mu, self.sd)

    @add.implementor(symmetric=True)
    def add(self, other: float):
        return (self._inner() + other).truncate(self.a + other, self.b + other)

    @mul.implementor(symmetric=True)
    def mul(self, other: float):
        a = self.a * other
        b = self.b * other
        if other < 0:
            a, b = b, a
        return (self._inner() * other).truncate(a, b)

    def truncate(self, a=None, b=None):
        a = max(a, self.a) if a is not None else self.a
        b = min(b, self.b) if b is not None else self.b
        if (a, b) == (self.a, self.b):
            return self

        return self._inner().truncate(a, b)

    def args_str(self):
        return self.mu, self.sd, self.a, self.b


BernoulliDistribution = rv_discrete_class(bernoulli, 'BernoulliDistribution', __name__,
                                          support_space_cls=DiscreteWholeSupportSpace)

ShiftedGeometricDistribution = rv_discrete_class(geom, 'ShiftedGeometricDistribution', __name__,
                                                 support_space_cls=DiscreteWholeSupportSpace)


class UnshiftedGeometricDistribution(ShiftedGeometricDistribution):
    def __init__(self, p):
        super().__init__(p, -1)
        self.p = p

    def args_str(self):
        return self.p

    def support_space(self):
        return super().support_space() + (-1)


HyperGeometricDistribution = rv_discrete_class(hypergeom, 'HyperGeometricDistribution', __name__,
                                               support_space_cls=DiscreteWholeSupportSpace)

PoissonDistribution = rv_discrete_class(poisson, 'PoissonDistribution', __name__,
                                        support_space_cls=DiscreteWholeSupportSpace)

ZipfDistribution = rv_discrete_class(zipf, 'ZipfDistribution', __name__,
                                     support_space_cls=DiscreteWholeSupportSpace)


class TriangularDistribution(rv_continuous_class(triang)):
    def __init__(self, a, b, peak):
        loc = a
        scale = b - loc
        c = (peak - loc) / scale
        super().__init__(c, loc, scale)
        self.a = a
        self.b = b
        self.peak = peak

    def args_str(self):
        return self.a, self.b, self.peak

    @add.implementor()
    def add(self, other: float):
        return type(self)(self.a + other, self.b + other, self.peak + other)

    @mul.implementor()
    def mul(self, other: float):
        a = self.a
        b = self.b
        if other < 0:
            a, b = b, a
        return type(self)(a * other, b * other, self.peak * other)



from snake_eyes import BernoulliDistribution, Distribution, NormalDistribution, UniformDistribution, \
    UniformDiscreteDistribution

radermacher_distribution: Distribution[int] = (2 * BernoulliDistribution(0.5) - 1)
standard_normal_distribution: Distribution[float] = NormalDistribution(0, 1)
standard_uniform_distribution: Distribution[float] = UniformDistribution(0, 1)


def die(sides, n=1) -> Distribution[int]:
    if n != 1:
        return n @ die(sides)
    return UniformDiscreteDistribution(1, sides)


coin: Distribution[int] = die(2)
d6: Distribution[int] = die(6)



from snake_eyes_data import __url__, __author__, __version__
from snake_eyes.distribution import Distribution, DistributionKind, ConstDistribution
from snake_eyes.with_rv import UniformDistribution, UniformDiscreteDistribution, NormalDistribution, \
    BernoulliDistribution, ShiftedGeometricDistribution, UnshiftedGeometricDistribution, HyperGeometricDistribution, \
    PoissonDistribution, ZipfDistribution
from snake_eyes.buffered_distribution import ChoiceDistribution
from snake_eyes.common_distributions import radermacher_distribution, standard_uniform_distribution, \
    standard_normal_distribution, coin, die, d6

from snake_eyes.additional_distributions.drop import drop_sum, drop_prod
from snake_eyes.additional_distributions.min_max import dmin, dmax
from snake_eyes.additional_distributions.split import SplitDistribution
from snake_eyes.additional_distributions.extra import explode



from typing import Generic, TypeVar, Iterable, Optional

import numpy as np

from snake_eyes.additional_distributions.split import JoinedSupportSpace
from snake_eyes.buffered_distribution import BufferedDistribution
from snake_eyes.bufferer import Bufferer
from snake_eyes.distribution import Distribution
from snake_eyes.support_space import SupportSpace
from snake_eyes.util import _maybe_parenthesise, prod

T = TypeVar('T')


def min_support_space(parts: Iterable[SupportSpace]):
    maxes = [p.maximum() for p in parts]
    partial_maxes = [
        min(m for (i, m) in maxes if i != j)
        for j, _ in enumerate(parts)
    ]
    return JoinedSupportSpace.join(
        p.truncate(None, m) for (p, m) in zip(parts, partial_maxes)
    )


class MinDistribution(Distribution[T], Generic[T]):
    def __init__(self, parts: Iterable[Distribution]):
        self.parts = tuple(parts)

    def get(self) -> T:
        return min(p.get() for p in self.parts)

    def cumulative_density(self, k) -> Optional[float]:
        inv_ret = 1
        for p in self.parts:
            cdf = p.cumulative_density(k)
            if cdf is None:
                return None
            inv_ret *= (1 - cdf)
        return 1 - inv_ret

    def support_space(self) -> Optional[SupportSpace]:
        ss = []
        for p in self.parts:
            s = p.support_space()
            if s is None:
                return None
            ss.append(s)
        return min_support_space(ss)

    def probability(self, k) -> Optional[float]:
        cdf = [
            p.cumulative_density(k) for p in self.parts
        ]
        prob = [
            p.probability(k) for p in self.parts
        ]
        if any(x is None for x in cdf) or any(x is None for x in prob):
            return None
        inv_cdf = [1 - x for x in cdf]
        prod_inv_cdf = prod(inv_cdf)
        return sum(
            (prod_inv_cdf / icdf) * p for (icdf, p) in zip(inv_cdf, prob)
        )

    def sample_cumulative_density(self, k, **kwargs) -> Optional[float]:
        inv_ret = 1
        for p in self.parts:
            cdf = p.approx_cumulative_density(k, **kwargs)
            inv_ret *= (1 - cdf)
        return 1 - inv_ret

    def sample_probability(self, k, sample_size=4096, epsilon=None) -> Optional[float]:
        cdf = [
            p.approx_cumulative_density(k, sample_size=sample_size) for p in self.parts
        ]
        prob = [
            p.approx_probability(k, sample_size=sample_size, epsilon=epsilon) for p in self.parts
        ]
        inv_cdf = [1 - x for x in cdf]
        prod_inv_cdf = prod(inv_cdf)
        return sum(
            (prod_inv_cdf / icdf) * p for (icdf, p) in zip(inv_cdf, prob)
        )

    def __str__(self):
        return type(self).__name__ + '(' \
               + ", ".join(_maybe_parenthesise(p) for p in self.parts) \
               + ')'

    def __eq__(self, other):
        return type(self) is type(other) \
               and set(self.parts) == set(self.parts)

    def __hash__(self):
        return hash((type(self), frozenset(self.parts)))


class MinBufferer(Bufferer):
    def __init__(self, parts: Iterable[Bufferer]):
        super().__init__()
        self.parts = tuple(parts)

    def get_buffer(self, size) -> np.ndarray:
        return np.min([p.get_buffer(size) for p in self.parts], axis=0)


class MinBufferedDistribution(BufferedDistribution[T], MinDistribution[T], Generic[T]):
    def __init__(self, parts: Iterable[BufferedDistribution[T]]):
        MinDistribution.__init__(self, parts)
        BufferedDistribution.__init__(self, MinBufferer(p.bufferer for p in parts))


def dmin(*parts: Distribution):
    if all(isinstance(p, BufferedDistribution) for p in parts):
        return MinBufferedDistribution(parts)
    return MinDistribution(parts)


def dmax(*parts: Distribution):
    return -dmin(*(-p for p in parts))



from __future__ import annotations

from abc import ABC, abstractmethod
from inspect import isabstract
from typing import Dict, FrozenSet, Tuple, Any, Iterable, Iterator

from dyndis import MultiDispatch, Self
import numpy as np

BUFFER_SIZE = 4096

empty_frozenset = frozenset()

add = MultiDispatch('__add__')
mul = MultiDispatch('__mul__')


class Bufferer(ABC, Iterator):
    """
    An infinite, self-filling buffer for non-deterministic functions
    """
    cache: Dict[Tuple[Tuple, FrozenSet[Tuple[Any, Any]]], Bufferer]

    @classmethod
    def make(cls, args, kwargs=None) -> Bufferer:
        """
        if the class cache contains a buffer of equal parameters, return it, if not create a new one and cache
         it for future use
        """
        fs = empty_frozenset if not kwargs else frozenset(kwargs.items())
        key = args, fs
        ret = cls.cache.get(key)
        if ret is None:
            ret = cls.cache[key] = cls(*args, **(kwargs or {}))
        return ret

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if not isabstract(cls):
            cls.cache = {}

    def __init__(self):
        self.buffer = []
        self.next_index = 0

    @abstractmethod
    def get_buffer(self, size) -> np.ndarray:
        """
        :return: a valid np array, buffer must be of exactly size `size`
        """
        pass

    def __next__(self):
        """
        :return: the next element in the generated buffer, generates a new buffer if necessary
        """
        if len(self.buffer) == self.next_index:
            self.buffer = self.get_buffer(BUFFER_SIZE)
            self.next_index = 1
            return self.buffer[0]
        ret = self.buffer[self.next_index]
        self.next_index += 1
        return ret

    def get_n(self, n) -> np.ndarray:
        """
        :return: an independent buffer of size n
        """

        if not len(self.buffer):
            self.buffer = self.get_buffer(max(n, BUFFER_SIZE))

        ret = np.empty(n, self.buffer.dtype)
        if self.next_index + n > len(self.buffer):
            first_part_end = len(self.buffer) - self.next_index
            ret[0:first_part_end] = self.buffer[self.next_index:]
            self.buffer = self.get_buffer(BUFFER_SIZE + n - first_part_end)
            ret[first_part_end:] = self.buffer[BUFFER_SIZE:]
            self.buffer.resize((BUFFER_SIZE,), refcheck=False)
            self.next_index = 0
        else:
            ret[:] = self.buffer[self.next_index: self.next_index + n]
            self.next_index += n

        return ret

    __radd__ = __add__ = add.op()
    __rmul__ = __mul__ = mul.op()

    def reciprocal(self):
        return InvBufferer.make((self,))

    def __neg__(self):
        return -1 * self

    def __sub__(self, other):
        return self + (-other)

    def __rsub__(self, other):
        return other + (-self)

    def __truediv__(self, other):
        if isinstance(other, Bufferer):
            return self * other.reciprocal()
        return self * (1 / other)

    def __rtruediv__(self, other):
        return other * self.reciprocal()

    @add.implementor(symmetric=True)
    def add(self, other: Any):
        if isinstance(other, Bufferer):
            return SumBufferer.make(((self, other),))
        if other == 0:
            return self
        return SumConstBufferer.make((self, other))

    @mul.implementor(symmetric=True)
    def mul(self, other: Any):
        if isinstance(other, Bufferer):
            return ProdBufferer.make(((self, other),))
        if other == 1:
            return self
        return ProdConstBufferer.make((self, other))


class FuncBufferer(Bufferer):
    """
    A bufferer that uses a function to generate buffers
    """

    def __init__(self, func, args, kwargs):
        super().__init__()
        self.args = args
        self.kwargs = kwargs
        self.func = func

    def get_buffer(self, size) -> np.ndarray:
        return self.func(*self.args, size=size, **self.kwargs)


def bufferer(func, name=None):
    class Ret(FuncBufferer):
        def __init__(self, *args, **kwargs):
            super().__init__(func, args, kwargs)

    Ret.__name__ = name or func.__name__
    return Ret


class SumBufferer(Bufferer):
    """
    A bufferer representing a sum of multiple bufferers
    """

    def __init__(self, parts: Iterable[Bufferer]):
        super().__init__()
        self.parts = tuple(parts)

    def get_buffer(self, size) -> np.ndarray:
        return np.sum([p.get_buffer(size) for p in self.parts], axis=0)

    def __add__(self, other):
        return type(self)(self.parts + (other,))

    def __radd__(self, other):
        return type(self)((other,) + self.parts)


class ProdBufferer(Bufferer):
    """
    A bufferer representing a product of multiple bufferers
    """

    def __init__(self, parts: Iterable[Bufferer]):
        super().__init__()
        self.parts = tuple(parts)

    def get_buffer(self, size) -> np.ndarray:
        return np.prod([p.get_buffer(size) for p in self.parts], axis=0)

    def __mul__(self, other):
        return type(self)(self.parts + (other,))

    def __rmul__(self, other):
        return type(self)((other,) + self.parts)


class InvBufferer(Bufferer):
    """
    A bufferer representing the inverse of a bufferer
    """

    def __init__(self, inner: Bufferer):
        super().__init__()
        self.inner = inner

    def get_buffer(self, size) -> np.ndarray:
        return 1 / self.inner.get_buffer(size)

    def reciprocal(self):
        return self.inner


class SumConstBufferer(Bufferer):
    """
    A bufferer representing a sum of a bufferers and a constant
    """

    def __init__(self, inner: Bufferer, const):
        super().__init__()
        self.inner = inner
        self.const = const

    def get_buffer(self, size) -> np.ndarray:
        return self.inner.get_buffer(size) + self.const

    @add.implementor(symmetric=True)
    def add(self, other: Any):
        if isinstance(other, Bufferer):
            return NotImplemented
        return self.inner + (self.const + other)

    @add.implementor()
    def add(self, other: Self):
        return (self.inner + other.inner) + (self.const + other.const)

    @add.implementor(symmetric=True)
    def add(self, other: SumBufferer):
        return (other + self.inner) + self.const


class ProdConstBufferer(Bufferer):
    """
    A bufferer representing a product of a bufferers and a constant
    """

    def __init__(self, inner: Bufferer, const):
        super().__init__()
        self.inner = inner
        self.const = const

    def get_buffer(self, size) -> np.ndarray:
        return self.inner.get_buffer(size) * self.const

    @mul.implementor(symmetric=True)
    def mul(self, other: Any):
        if isinstance(other, Bufferer):
            return NotImplemented
        return self.inner * (self.const * other)

    @mul.implementor()
    def mul(self, other: Self):
        return (self.inner * other.inner) * (self.const * other.const)

    @mul.implementor(symmetric=True)
    def mul(self, other: ProdBufferer):
        return (other * self.inner) * self.const


@bufferer
def ChoiceBufferer(options, p=None, *, size):
    return np.random.choice(options, p=p, size=size)



from __future__ import annotations

from itertools import chain
from numbers import Number
from typing import Generic, TypeVar, Iterable, Type, Any, Tuple, Dict, Optional, Mapping, Union

from dyndis import Self
import numpy as np

from snake_eyes.support_space import DiscreteFiniteSupportSpace

try:
    from scipy import stats
except ImportError:
    stats = None

from snake_eyes.bufferer import Bufferer, ChoiceBufferer
from snake_eyes.distribution import Distribution, add, mul, div, ConstDistribution, ReciprocalDistribution, \
    SumDistribution, ProductDistribution, _maybe_parenthesise
from snake_eyes.util import prod

T = TypeVar('T')


class BufferedDistribution( Distribution[T], Generic[T]):
    """
    A generic distribution that takes a bufferer and adapts it into a distribution
    """

    def __init__(self, bufferer: Bufferer):
        self.bufferer = bufferer

    def get(self) -> T:
        return next(self.bufferer)

    def reciprocal(self):
        return ReciprocalBufferedDistribution(self)

    @add.implementor()
    def add(self, other: Self):
        return SumBufferedDistribution((self, other))

    @add.implementor()
    def add(self, other: Number):
        return SumConstBufferedDistribution(self, other)

    @add.implementor()
    def add(self, other: ConstDistribution):
        return self + other.value

    @mul.implementor()
    def mul(self, other: Self):
        return ProductBufferedDistribution((self, other))

    @mul.implementor()
    def mul(self, other: Number):
        return ProductConstBufferedDistribution(self, other)

    @mul.implementor()
    def mul(self, other: ConstDistribution):
        return self * other.value

    def get_n(self, n):
        return self.bufferer.get_n(n)


class ReciprocalBufferedDistribution(BufferedDistribution[T], ReciprocalDistribution, Generic[T]):
    def __init__(self, inner: BufferedDistribution[T]):
        BufferedDistribution.__init__(self, inner.bufferer.reciprocal())
        ReciprocalDistribution.__init__(self, inner)


class SumBufferedDistribution(BufferedDistribution[T], SumDistribution, Generic[T]):
    def __init__(self, parts: Iterable[BufferedDistribution[T]]):
        SumDistribution.__init__(self, parts)
        BufferedDistribution.__init__(self, sum(p.bufferer for p in self.parts))

    @add.implementor(symmetric=True)
    def add(self, other: BufferedDistribution):
        p = self.parts + (other,)
        return type(self)(p)


class SumConstBufferedDistribution(BufferedDistribution[T], Generic[T]):
    """
    A distribution that is the sum of a buffered distribution and a constant value
    """

    def __init__(self, inner: BufferedDistribution[T], const: T):
        self.inner = inner
        self.const = const
        super().__init__(inner.bufferer + const)

    def mean(self):
        m = self.inner.mean()
        if m is None:
            return None
        return m + self.const

    def variance(self):
        return self.inner.variance()

    def cumulative_density(self, k):
        return self.inner.cumulative_density(k - self.const)

    def probability(self, k):
        return self.inner.probability(k - self.const)

    def support_space(self):
        iss = self.inner.support_space()
        return iss and (iss + self.const)

    @add.implementor(symmetric=True)
    def add(self, other: BufferedDistribution):
        return (self.inner + other) + self.const

    @add.implementor(symmetric=True)
    def add(self, other: Any):
        if isinstance(other, Distribution):
            return NotImplemented
        return self.inner + (self.const + other)

    @add.implementor()
    def add(self, other):
        return (self.inner + other.inner) + (self.const + other.const)

    @add.implementor(symmetric=True)
    def add(self, other: SumBufferedDistribution):
        return (other + self.inner) + self.const

    @mul.implementor(symmetric=True)
    def mul(self, other: Any):
        return self.inner * other + self.const * other

    @div.implementor()
    def truediv(self, other: Any):
        return self.inner / other + self.const / other

    def __eq__(self, other):
        return type(self) is type(other) and (self.inner, self.const) == (other.inner, other.const)

    def __str__(self):
        return f'{_maybe_parenthesise(self.inner)} + {self.const}'

    def __hash__(self):
        return hash((type(self), self.inner, self.const))


class ProductBufferedDistribution( BufferedDistribution[T], ProductDistribution, Generic[T]):
    def __init__(self, parts: Iterable[BufferedDistribution[T]]):
        ProductDistribution.__init__(self, parts)
        BufferedDistribution.__init__(self, prod(p.bufferer for p in self.parts))

    @mul.implementor(symmetric=True)
    def mul(self, other: BufferedDistribution):
        p = self.parts + (other,)
        return type(self)(p)


class ProductConstBufferedDistribution( BufferedDistribution[T], Generic[T]):
    """
    A distribution that is the product of a buffered distribution and a constant value
    """

    def __init__(self, inner: BufferedDistribution[T], const: T):
        self.inner = inner
        self.const = const
        super().__init__(inner.bufferer * const)

    def mean(self):
        m = self.inner.mean()
        if m is None:
            return None
        return m * self.const

    def variance(self):
        m = self.inner.variance()
        if m is None:
            return None
        return m * self.const ** 2

    def support_space(self):
        iss = self.inner.support_space()
        return iss and (iss * self.const)

    def cumulative_density(self, k):
        return self.inner.cumulative_density(k / self.const)

    def probability(self, k):
        return self.inner.probability(k / self.const)

    @mul.implementor(symmetric=True)
    def mul(self, other: BufferedDistribution):
        return (self.inner * other) * self.const

    @mul.implementor(symmetric=True)
    def mul(self, other: Any):
        if isinstance(other, Distribution):
            return NotImplemented
        return self.inner * (self.const * other)

    @mul.implementor()
    def mul(self, other):
        return (self.inner * other.inner) * (self.const * other.const)

    @mul.implementor(symmetric=True)
    def mul(self, other: ProductBufferedDistribution):
        return (other * self.inner) * self.const

    def __eq__(self, other):
        return type(self) is type(other) and (self.inner, self.const) == (other.inner, other.const)

    def __str__(self):
        return f'{_maybe_parenthesise(self.inner)} * {self.const}'

    def __hash__(self):
        return hash((type(self), self.inner, self.const))


class BuffererMakerDistribution( BufferedDistribution[T], Generic[T]):
    """
    A specialized bufferer distribution that makes use of already created and cached bufferers using the
     bufferer's make method.
    """

    def __init__(self, bufferer_cls: Type[Bufferer], args, kwargs=None):
        super().__init__(bufferer_cls.make(args, kwargs))
        self.args: Tuple[Tuple, Optional[Dict[str, Any]]] = (args, kwargs)

    def __repr__(self):
        args_str = self.args_str()
        args = ()
        kwargs = None
        if isinstance(args_str, Mapping):
            kwargs = args_str
        elif len(args_str) != 2 or not isinstance(args_str[1], Mapping) or not isinstance(args_str[0], Iterable):
            args = args_str
        else:
            args, kwargs = args_str

        args_str = (repr(a) for a in args)
        if kwargs:
            args_str = chain(args_str, (f'{k}={v!r}' for (k, v) in kwargs.items()))
        return type(self).__name__ + "(" + ", ".join(args_str) + ")"

    def args_str(self) -> Union[Tuple[Iterable, Optional[Mapping[str, Any]]], Iterable, Mapping[str, Any]]:
        return self.args

    def __eq__(self, other):
        return type(self) is type(other) and self.args == other.args

    def __hash__(self):
        return hash(repr(self))


class ChoiceDistribution(BuffererMakerDistribution):
    """
    A discrete distribution that chooses from a numpy array as np.random.choice
    """

    def __init__(self, choices, p=None):
        choices = tuple(choices)
        if p is not None:
            p = tuple(p)
        super().__init__(ChoiceBufferer, (choices,), {'p': p})
        self.choices = choices
        self.p = p

    def mean(self):
        if self.p is not None:
            return sum(
                i * p for (i, p) in zip(self.choices, self.p)
            )
        return sum(
            i for i in self.choices
        ) / len(self.choices)

    def variance(self):
        if self.p is not None:
            return sum(
                i * p for (i, p) in zip(self.choices, self.p)
            ) - self.mean()
        return sum(
            i for i in self.choices
        ) / len(self.choices) - self.mean()

    def support_space(self):
        return DiscreteFiniteSupportSpace(self.choices)

    def cumulative_density(self, k):
        if self.p is not None:
            return np.sum(self.p[self.choices <= k])
        return np.sum(self.choices <= k) / len(self.choices)

    def probability(self, k):
        if self.p is not None:
            return np.sum(self.p[self.choices == k])
        return np.sum(self.choices == k) / len(self.choices)

    @add.implementor(symmetric=True)
    def add(self, other: Number):
        choices = [c + other for c in self.choices]
        return type(self)(choices, p=self.p)

    @mul.implementor(symmetric=True)
    def mul(self, other: Number):
        choices = [c * other for c in self.choices]
        return type(self)(choices, p=self.p)

    def reciprocal(self):
        choices = [1 / c for c in self.choices]
        return type(self)(choices, p=self.p)

    def truncate(self, min=None, max=None):
        if self.p is not None:
            choices = []
            probs = []
            prob_sum = 0
            for c, p in zip(self.choices, self.p):
                if (min is None or min <= c) and (max is None or max >= c):
                    choices.append(c)
                    probs.append(p)
                    prob_sum += p
            if not prob_sum:
                raise ValueError("can't truncate all options")
            probs = [p / prob_sum for p in probs]
            return type(self)(choices, p=probs)
        else:
            choices = []
            for c in self.choices:
                if (min is None or min <= c) and (max is None or max >= c):
                    choices.append(c)
            if not choices:
                raise ValueError("can't truncate all options")
            return type(self)(choices)



from __future__ import annotations

from abc import abstractmethod, ABC
from bisect import bisect_right
from enum import Enum, auto
from itertools import product, count, chain
from math import sqrt
from numbers import Rational
from typing import Sequence, Optional, Iterable, Collection

from containerview.cached_iter import CacheIter
from containerview.sequence_view import SequenceView

from snake_eyes.util import prod, min_max_prod


class DistributionKind(Enum):
    """
    Describes whether a distribution kind is
    """
    Discrete = auto()
    # if a continuous support space includes values (a,b), it must also include all values between them
    Continuous = auto()


class SupportSpace(ABC):
    @abstractmethod
    def minimum(self):
        """
        :return: the smallest possible value the support can return, if different than -float('inf'),
         must be a value that the support includes
        """
        pass

    @abstractmethod
    def maximum(self):
        """
        :return: the largest possible value the support can return, if different than float('inf'),
         must be a value that the support includes
        """
        pass

    @abstractmethod
    def contains(self, k) -> Optional[bool]:
        """
        :return: whether the support space can return the value, or None if that can't be calculated
        """
        if self.minimum() > k or self.maximum() < k:
            return False
        if self.kind() == DistributionKind.Continuous:
            return True
        return None

    @abstractmethod
    def kind(self) -> Optional[DistributionKind]:
        """
        :return: the kind (discrete or continuous) of the support, or None if it is neither
        """
        pass

    @abstractmethod
    def is_finite(self):
        """
        :return: whether the support includes a finite number of values
        """
        pass

    def __iter__(self):
        """
        iterate over all possible values of the support, by default, raises error if the support is not discrete
        """
        if self.kind() != DistributionKind.Discrete:
            raise ValueError('support space is non-discrete')
        yield from ()

    def __add__(self, other: Rational):
        return SumSupportSpace((self, DiscreteFiniteSupportSpace((other,), True)))

    def __mul__(self, other: Rational):
        return ProdSupportSpace((self, DiscreteFiniteSupportSpace((other,), True)))

    def truncate(self, a, b) -> Optional[SupportSpace]:
        """
        truncate the support to exclude certain values
        """
        if a is None:
            a = -float('inf')
        if b is None:
            b = float('inf')

        if a <= self.minimum() and b >= self.maximum():
            return self
        if a > self.maximum() or b < self.minimum() or a > b:
            return None
        return TruncatedSupportSpace(self, a, b)

    def contains_space(self, other: SupportSpace):
        """
        :return: whether self wholly includes all values in other
        """
        s_k = self.kind()
        o_k = other.kind()
        if o_k == DistributionKind.Discrete and other.is_finite():
            return all(self.contains(t) for t in other)
        elif o_k == DistributionKind.Continuous:
            if s_k is None:
                return None
            if s_k != DistributionKind.Continuous:
                return False
            return self.contains(other.minimum()) and self.contains(other.maximum())
        else:
            return None

    def __lt__(self, other):
        return self.maximum() < other.minimum()

    def __le__(self, other):
        return self.maximum() <= other.minimum()


class DiscreteFiniteSupportSpace(SupportSpace):
    """
    A discrete, finite support space
    """

    def __init__(self, support: Sequence, is_sorted=False):
        self.support = support
        self.sorted = is_sorted

    def minimum(self):
        if self.sorted:
            return self.support[0]
        return min(self.support)

    def maximum(self):
        if self.sorted:
            return self.support[-1]
        return max(self.support)

    def contains(self, k) -> bool:
        if self.sorted:
            bis_ind = bisect_right(self.support, k)
            return self.support[bis_ind] == k
        return k in self.support

    def kind(self) -> DistributionKind:
        return DistributionKind.Discrete

    def is_finite(self):
        return True

    def __iter__(self):
        yield from super().__iter__()
        yield from self.support

    def __add__(self, other):
        return type(self)([s + other for s in self.support], self.sorted)

    def __mul__(self, other):
        return type(self)([s * other for s in self.support], self.sorted)

    def truncate(self, a, b):
        if a is None:
            a = -float('inf')
        if b is None:
            b = float('inf')

        if self.sorted and self.minimum() <= a and self.maximum() >= b:
            return self

        if self.sorted and (self.minimum() > b or self.maximum() < b):
            return None

        ss = tuple(s for s in self.support if a <= s <= b)
        if not ss:
            return None

        return type(self)(
            ss,
            self.sorted
        )


class DiscreteWholeSupportSpace(SupportSpace):
    """
    A discrete support space that only supports whole values in its range
    """

    def __init__(self, min, max=None):
        if min == -float('inf'):
            min = None
        elif min is not None:
            min = int(min)

        if max == float('inf'):
            max = None
        elif max is not None:
            max = int(max)

        self.min = min
        self.max = max

    def minimum(self):
        if self.min is not None:
            return self.min
        return -float('inf')

    def maximum(self):
        if self.max is not None:
            return self.max
        return float('inf')

    def contains(self, k) -> bool:
        return k % 1 == 0 \
               and (self.min is None or self.min <= k) \
               and (self.max is None or self.max >= k)

    def kind(self) -> Optional[DistributionKind]:
        return DistributionKind.Discrete

    def is_finite(self):
        return self.min is not None \
               and self.max is not None

    def __iter__(self):
        if self.min is None:
            if self.max is None:
                yield 0
                for i in count(1):
                    yield i
                    yield -i
            else:
                yield from count(self.max, -1)
        else:
            if self.max is None:
                yield from count(self.min)
            else:
                yield from range(self.min, self.max + 1)

    def __add__(self, other):
        n = None if self.min is None else self.min + other
        x = None if self.max is None else self.max + other
        return type(self)(n, x)

    def __mul__(self, other):
        n = None if self.min is None else self.min * other
        x = None if self.max is None else self.max * other
        if other < 0:
            n, x = x, n
        return type(self)(n, x)

    def truncate(self, a, b):
        if a is None:
            a = self.min
        elif self.min is not None:
            a = max(self.min, a)

        if b is None:
            b = self.max
        elif self.min is not None:
            b = min(self.max, b)

        if b < a:
            return None

        return type(self)(a, b)


class ContinuousSupportSpace(SupportSpace):
    """
    A continuous support space between two bounds
    """

    def __init__(self, minimum=-float('inf'), maximum=float('inf')):
        self._minimum = minimum
        self._maximum = maximum

    def minimum(self):
        return self._minimum

    def maximum(self):
        return self._maximum

    def contains(self, k) -> bool:
        return self._minimum <= k <= self._maximum

    def kind(self) -> DistributionKind:
        return DistributionKind.Continuous

    def __add__(self, other):
        return type(self)(self._minimum + other, self._maximum + other)

    def __mul__(self, other):
        n = self._minimum * other
        x = self._maximum * other
        if other < 0:
            n, x = x, n
        return type(self)(n, x)

    def truncate(self, a, b):
        if a is None:
            a = -float('inf')
        if b is None:
            b = float('inf')

        if self.minimum() <= a and self.maximum() >= b:
            return self

        a = max(a, self._minimum)
        b = min(b, self._maximum)

        if b < a:
            return None

        return type(self)(a, b)

    def is_finite(self):
        return False


class CompundSupportSpace(SupportSpace, ABC):
    def __init__(self, parts: Collection[SupportSpace]):
        self.parts = parts

    def is_finite(self):
        return all(p.is_finite() for p in self.parts)

    def __iter__(self):
        if any(not p.is_finite() for p in self.parts):
            # cantor tupling
            def _unpack(buffer, index):
                if len(self.parts) - len(buffer) == 1:
                    buffer.append(index)
                    return
                w = int((sqrt(8 * index + 1) - 1) // 2)
                buffer.append(index - (w ** 2 + w) // 2)
                _unpack(buffer, (w ** 2 + 3 * w) // 2 - index)

            def unpack(index):
                buffer = []
                _unpack(buffer, index)
                buffer.reverse()
                return buffer

            cached = [CacheIter(iter(p)) for p in self.parts]
            for i in count():
                indices = unpack(i)
                yield tuple(cache[i] for (cache, i) in zip(cached, indices))
        else:
            yield from product(*self.parts)


class ZipSupportSpace(CompundSupportSpace):
    """
    A zipping of support spaces
    """

    def minimum(self):
        return tuple(p.minimum() for p in self.parts)

    def maximum(self):
        return tuple(p.maximum() for p in self.parts)

    def contains(self, k) -> bool:
        return len(k) == len(self.parts) \
               and all(p.contains(ki) for (p, ki) in zip(self.parts, k))

    def kind(self) -> Optional[DistributionKind]:
        ret = None
        for p in self.parts:
            k = p.kind()
            if k is None:
                return k
            if ret is None:
                ret = k
            elif ret != k:
                return None
        return ret

    def __add__(self, other):
        return type(self)(p + other for p in self.parts)

    def __mul__(self, other):
        return type(self)(p * other for p in self.parts)


class SymmetricOpSupportSpace(CompundSupportSpace, ABC):
    """
    an abstract support space for combining support spaces
    """

    @classmethod
    @abstractmethod
    def func(cls, args: Iterable):
        pass

    @classmethod
    @abstractmethod
    def unfunc(cls, targ, known):
        pass

    @classmethod
    def _contains_from(cls, parts: SequenceView, targ):
        part = parts[0]
        parts = parts[1:]
        if not parts:
            return part.contains(targ)
        for s in part:
            new_targ = cls.unfunc(targ, s)
            if cls._contains_from(parts, new_targ):
                return True
        return False

    def contains(self, k):
        try:
            return self._contains_from(SequenceView(self.parts), k)
        except ValueError:
            return super().contains(k)

    def __iter__(self):
        yield from (self.func(t) for t in super().__iter__())

    def kind(self):
        ret = None
        for p in self.parts:
            k = p.kind()
            if k is None:
                return k
            if ret is None or k == DistributionKind.Continuous:
                ret = k
        return ret


class SumSupportSpace(SymmetricOpSupportSpace):
    """
    Compound sum of support spaces
    """

    @classmethod
    def func(cls, args):
        return sum(args)

    @classmethod
    def unfunc(cls, targ, k):
        return targ - k

    def minimum(self):
        return self.func(p.minimum() for p in self.parts)

    def maximum(self):
        return self.func(p.maximum() for p in self.parts)

    def __add__(self, other):
        parts = list(self.parts)
        parts[0] += other
        return type(self)(parts)


class ProdSupportSpace(SymmetricOpSupportSpace):
    """
    Compound product of support spaces
    """

    @classmethod
    def func(cls, args):
        return prod(args)

    @classmethod
    def unfunc(cls, targ, k):
        return targ / k

    def _minmax(self):
        mins = tuple(p.minimum() for p in self.parts)
        maxs = tuple(p.maximum() for p in self.parts)
        if any(m is None for m in mins) or any(m is None for m in maxs):
            return None, None
        return min_max_prod(mins, maxs)

    def minimum(self):
        return self._minmax()[0]

    def maximum(self):
        return self._minmax()[1]

    def __mul__(self, other):
        parts = list(self.parts)
        parts[0] *= other
        return type(self)(parts)


class ReciprocalSupportSpace(SupportSpace):
    """
    An inverse of a support space
    """

    def __init__(self, inner: SupportSpace):
        self.inner = inner

    def minimum(self):
        i_min = self.inner.minimum()
        i_max = self.inner.maximum()
        if (i_min < 0) != (i_max < 0):
            return -float('inf')
        return 1 / i_max

    def maximum(self):
        i_min = self.inner.minimum()
        i_max = self.inner.maximum()
        if (i_min < 0) != (i_max < 0):
            return float('inf')
        return 1 / i_min

    def kind(self) -> Optional[DistributionKind]:
        return self.inner.kind()

    def contains(self, k) -> bool:
        return self.inner.contains(1 / k)

    def __iter__(self):
        yield from (1 / v for v in self.inner)

    def is_finite(self):
        return self.inner.is_finite()


class TruncatedSupportSpace(SupportSpace):
    """
    a truncated support space
    """

    def __init__(self, inner, a, b):
        self.inner = inner
        if a is None:
            a = -float('inf')
        if b is None:
            b = float('inf')
        self.a = a
        self.b = b

    def minimum(self):
        return min(self.inner.minimum(), self.a)

    def maximum(self):
        return max(self.inner.maximum(), self.b)

    def contains(self, k) -> bool:
        return self.a <= k <= self.b and self.inner.contains(k)

    def kind(self) -> Optional[DistributionKind]:
        return self.inner.kind()

    def __iter__(self):
        yield from (
            k for k in self.inner
            if self.a <= k <= self.b
        )

    def __add__(self, other):
        return (self.inner + other).truncate(self.a + other, self.b + other)

    def __mul__(self, other):
        a = self.a * other
        b = self.b * other
        if other < 0:
            a, b = b, a
        return (self.inner * other).truncate(a, b)

    def truncate(self, a, b):
        if a is None:
            a = -float('inf')
        if b is None:
            b = float('inf')

        a, b = max(a, self.a), min(b, self.b)

        if b < a:
            return None

        return self.inner.truncate(a, b)

    def is_finite(self):
        return self.inner.is_finite()



from typing import TypeVar, Callable, Union, Container

from snake_eyes.support_space import DistributionKind
from snake_eyes.distribution import Distribution

T = TypeVar('T')


def explode(inner: Distribution[T], exploding_func: Union[Callable[[T], bool], Container[T]] = ...):
    if exploding_func is ...:
        ss = inner.support_space()
        if ss is None or ss.kind() != DistributionKind.Discrete:
            raise Exception('exploding_func must be declared for non-discrete distributions')
        m = ss.maximum()

        def exploding_func(x):
            return x == m
    elif isinstance(exploding_func, Callable):
        pass
    elif isinstance(exploding_func, Container):
        exploding_func = exploding_func.__contains__
    else:
        raise TypeError(f'cannot parse exploding_func {exploding_func}')

    ret = inner.map(
        lambda x: x + ret.get() if exploding_func(x) else x
    )
    return ret



import setuptools

import snake_eyes_data as snake_eyes

setuptools.setup(
    name='snake_eyes',
    version=snake_eyes.__version__,
    url=snake_eyes.__url__,
    author=snake_eyes.__author__,
    packages=['snake_eyes', 'snake_eyes_data', 'snake_eyes.additional_distributions'],
    python_requires='>=3.7.0',
    include_package_data=True,
    data_files=[
        ('', ['README.md', 'CHANGELOG.md']),
    ],
    install_requires=["dyndis", "numpy", 'scipy', 'containerview']
)



import numpy as np

try:
    from math import prod
except ImportError:
    from functools import reduce
    import operator as op


    def prod(x):
        """
        implementation of math.prod for python < 3.8
        """
        return reduce(op.mul, x)


def min_max_prod(minimums, maximums):
    """
    returns the smallest and largest products achievable by interweaving maximums and minimums

    >>> min_max_prod((3,6,0,1,6),(-4,3,-5,-4,0))
    (-2880, 2160)
    """
    assert len(maximums) == len(minimums)

    ret_min, ret_max = 1, 1
    for op_min, op_max in zip(minimums, maximums):
        cands = (op_min * ret_min, op_min * ret_max, op_max * ret_min, op_max * ret_max)
        ret_min = min(cands)
        ret_max = max(cands)
    return ret_min, ret_max


def _maybe_parenthesise(x, convert=str):
    """
    convert an object to string and surround in parenthesis if it contains an operator
    """
    warrenters = '+*/\\-^'
    ret = convert(x)
    if any((w in ret) for w in warrenters):
        return "(" + ret + ")"
    return ret


def common_dtype(*args):
    """
    Get the common dtype for numpy arrays, or object if none found
    """
    try:
        return np.result_type(*args)
    except TypeError:
        return object



from itertools import chain
from numbers import Rational
from typing import Generic, TypeVar, Iterable, Tuple, Any, Union, Optional

import numpy as np

from snake_eyes.support_space import SupportSpace, DistributionKind
from snake_eyes.buffered_distribution import ChoiceDistribution
from snake_eyes.distribution import Distribution, add, mul
from snake_eyes.util import common_dtype, _maybe_parenthesise

T = TypeVar('T')


class JoinedSupportSpace(SupportSpace):
    """
    Compound joining of support spaces
    """

    def __init__(self, parts: Iterable[SupportSpace]):
        self.parts = parts

    def kind(self) -> Optional[DistributionKind]:
        ret = None
        for p in self.parts:
            k = p.kind()
            if k is None or k == DistributionKind.Continuous:
                return None
            if ret is None:
                ret = k
            elif ret != k:
                return None
        return ret

    def truncate(self, a, b):
        parts_left = []
        for p in self.parts:
            parts_left.append(p.truncate(a, b))
        if all(p is None for p in parts_left):
            return None
        return type(self)(parts_left)

    def is_finite(self):
        return all(p.is_finite() for p in self.parts)

    @classmethod
    def join(cls, parts):
        uncontained = []
        parts = list(parts)
        while parts:
            part = parts.pop()
            if part is None:
                continue
            if all(not p.contains_space(part) for p in parts):
                uncontained.append(part)
        return cls(parts)

    def minimum(self):
        return min(p.minimum() for p in self.parts)

    def maximum(self):
        return max(p.maximum() for p in self.parts)

    def contains(self, k) -> bool:
        return any(p.contains(k) for p in self.parts)

    def __add__(self, other: Rational):
        return type(self)(tuple(p + other for p in self.parts))

    def __mul__(self, other: Rational):
        return type(self)(tuple(p * other for p in self.parts))

    def __iter__(self):
        infinite = []
        non_discrete = []
        continuous = []
        for p in self.parts:
            if p.is_finite():
                yield from p
            elif p.kind() == DistributionKind.Continuous:
                continuous.append(p)
            elif p.kind() == DistributionKind.Discrete:
                infinite.append(p)
            else:
                non_discrete.append(p)
        yield from chain.from_iterable(infinite)
        yield from chain.from_iterable(non_discrete)
        yield from chain.from_iterable(continuous)


class SplitDistribution(Distribution[T], Generic[T]):
    """
    A distribution that has a chance to delegate other distributions
    """

    def __init__(self, options: Iterable[Distribution[T]], probabilites: Iterable[float]):
        self.options: Tuple[Distribution[T]] = tuple(options)
        self.probabilities = tuple(probabilites)
        self.index_distribution = ChoiceDistribution(range(len(self.options)), p=self.probabilities)

    @classmethod
    def create(cls, options: Iterable[Tuple[Any, Union[type(...), float]]], normalize=False):
        partial_total = 0
        dists = []
        probs = []
        for (d, p) in options:
            dists.append(cls.coerce(d))
            if p is ...:
                if normalize:
                    raise ValueError('cannot use normalize and ... together')
                p = 1 - partial_total
            if p == 0:
                continue
            if p < 0:
                raise ValueError(f'option {d} had negative probability')
            probs.append(p)
            partial_total += p
            if not normalize and partial_total > 1:
                raise ValueError('total probabilities exceed 1')
        if normalize:
            probs = (p / partial_total for p in probs)
        return cls(dists, probs)

    def get(self):
        return self.options[self.index_distribution.get()].get()

    def get_n(self, n):
        indices = np.asanyarray(self.index_distribution.get_n(n))
        individuals = [
            np.asanyarray(option.get_n(np.sum(indices == i)))
            for (i, option) in enumerate(self.options)
        ]
        dtype = common_dtype(*individuals)
        ret = np.empty(shape=n, dtype=dtype)
        for i in range(len(individuals)):
            ret[indices == i] = individuals[i]
        return ret

    def mean(self):
        ret = 0
        for o, p in zip(self.options, self.probabilities):
            m = o.mean()
            if m is None:
                return None
            ret += (m * p)
        return ret

    def variance(self):
        mean = self.mean()
        if mean is None:
            return None
        ret = mean ** 2
        for p, sub_dist in zip(self.probabilities, self.options):
            subdist_mean = sub_dist.mean()
            if subdist_mean is None:
                return None
            subdist_var = sub_dist.variance()
            if subdist_var is None:
                return None
            subdist_mean_sq = subdist_var + subdist_mean ** 2
            ret += p * (subdist_mean_sq - 2 * mean * subdist_mean)

        return ret

    def support_space(self):
        ss = tuple(p.support_space() for p in self.options)
        if None in ss:
            return None
        return JoinedSupportSpace.join(ss)

    def cumulative_density(self, k):
        ret = 0
        for (o, p) in zip(self.options, self.probabilities):
            m = o.cumulative_density(k)
            if m is None:
                return None
            ret += p * m
        return ret

    def probability(self, k):
        ret = 0
        for (o, p) in zip(self.options, self.probabilities):
            m = o.probability(k)
            if m is None:
                return None
            ret += p * m
        return ret

    def __neg__(self):
        return type(self)(
            (-s for s in self.options),
            self.probabilities
        )

    def reciprocal(self):
        return type(self)(
            (s.reciprocal() for s in self.options),
            self.probabilities
        )

    @add.implementor(symmetric=True)
    def add(self, other: Any):
        return type(self)(
            (s + other for s in self.options),
            self.probabilities
        )

    @mul.implementor(symmetric=True)
    def mul(self, other: Any):
        return type(self)(
            (s * other for s in self.options),
            self.probabilities
        )

    def truncate(self, min=None, max=None):
        return type(self)(
            (s.truncate(min, max) for s in self.options),
            self.probabilities
        )

    def __str__(self):
        return type(self).__name__ + '(' \
               + ", ".join(f"{p:.1%}: {_maybe_parenthesise(c)}" for (c, p) in zip(self.options, self.probabilities)) \
               + ')'

    def __eq__(self, other):
        return type(self) is type(other) \
               and set(zip(self.options, self.probabilities)) == set(zip(other.options, other.probabilites))

    def __hash__(self):
        return hash((type(self), frozenset(zip(self.options, self.probabilities))))
